import numpy as np
import pandas as pd

from woodwork.statistics_utils.frequency_inference._types import DataCheckMessageCode

# https://pandas.pydata.org/docs/user_guide/timeseries.html#offset-aliases
ALL_ALIASES = [
    {"alias": ["B"], "desc": "business day frequency"},
    {"alias": ["C"], "desc": "custom business day frequency"},
    {"alias": ["D"], "desc": "calendar day frequency"},
    {"alias": ["W-SUN", "W"], "desc": "weekly frequency"},
    {"alias": ["M"], "desc": "month end frequency"},
    {"alias": ["SM"], "desc": "semi-month end frequency (15th and end of month)"},
    {"alias": ["BM"], "desc": "business month end frequency"},
    {"alias": ["CBM"], "desc": "custom business month end frequency"},
    {"alias": ["MS"], "desc": "month start frequency"},
    {"alias": ["SMS"], "desc": "semi-month start frequency (1st and 15th)"},
    {"alias": ["BMS"], "desc": "business month start frequency"},
    {"alias": ["CBMS"], "desc": "custom business month start frequency"},
    {"alias": ["Q-DEC", "Q"], "desc": "quarter end frequency"},
    {"alias": ["BQ-DEC", "BQ"], "desc": "business quarter end frequency"},
    {"alias": ["QS-OCT", "QS"], "desc": "quarter start frequency"},
    {"alias": ["BQS-OCT", "BQS"], "desc": "business quarter start frequency"},
    {"alias": ["A-DEC", "A", "Y"], "desc": "year end frequency"},
    {"alias": ["BA-DEC", "BA", "BY"], "desc": "business year end frequency"},
    {"alias": ["AS-JAN", "AS", "YS"], "desc": "year start frequency"},
    {"alias": ["BAS-JAN", "BAS", "BYS"], "desc": "business year start frequency"},
    {"alias": ["BH"], "desc": "business hour frequency"},
    {"alias": ["H"], "desc": "hourly frequency"},
    {"alias": ["T", "min"], "desc": "minutely frequency"},
    {"alias": ["S"], "desc": "secondly frequency"},
    {"alias": ["L", "ms"], "desc": "milliseconds"},
    {"alias": ["U", "us"], "desc": "microseconds"},
    {"alias": ["N"], "desc": "nanoseconds"},
]

# These keys are inferred to these values by pandas
KNOWN_FREQ_ISSUES = {
    "C": "B",
    "SM": None,
    "CBM": "BM",
    "SMS": None,
    "CBMS": "BMS",
}

HEAD_RANGE_LEN = 100
TAIL_RANGE_LEN = 100


def pad_datetime_series(dates, freq, pad_start=0, pad_end=100):
    dates = [pd.Timestamp(d) for d in dates]

    head = pd.Series([])
    tail = pd.Series([])

    if pad_start > 0:
        head = (
            pd.date_range(end=dates[0], periods=pad_start, freq=freq)[:-1]
        ).to_series()

    if pad_end > 0:
        tail = (
            pd.date_range(start=dates[-1], periods=pad_end, freq=freq)[1:]
        ).to_series()

    return (
        pd.concat([head, pd.Series(dates), tail])
        .reset_index(drop=True)
        .astype("datetime64[ns]")
    )


def generate_pandas_inferrable():
    dates = ["2005-01-01T00:00:00Z"]

    output = []
    for alias_obj in ALL_ALIASES:
        for freq in alias_obj["alias"]:
            pd_inferred_freq = (
                KNOWN_FREQ_ISSUES[freq]
                if freq in KNOWN_FREQ_ISSUES
                else alias_obj["alias"][0]
            )
            if pd_inferred_freq is not None:
                output.append(
                    {
                        "expected_infer_freq": pd_inferred_freq,
                        "dates": pad_datetime_series(
                            dates, freq=freq, pad_end=TAIL_RANGE_LEN
                        )[1:],
                    }
                )

    return output


def generate_infer_error_messages():
    cases = []
    dt1 = pd.date_range(end="2005-01-01", freq="H", periods=10)
    dt2 = pd.date_range(start=dt1[-1], freq="M", periods=10)[1:]
    dt3 = pd.date_range(start=dt2[-1], freq="D", periods=10)[1:]

    dates = dt1.append(dt2).append(dt3).to_series().reset_index(drop=True)

    cases.append(
        {
            "dates": dates,
            "expected_debug_obj": {
                "actual_range_start": dates.loc[0].isoformat(),
                "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
                "message": DataCheckMessageCode.DATETIME_SERIES_FREQ_CANNOT_BE_ESTIMATED,
                "estimated_freq": None,
                "estimated_range_start": None,
                "estimated_range_end": None,
                "duplicate_values": [],
                "missing_values": [],
                "extra_values": [],
                "nan_values": [],
            },
        }
    )

    dt1 = pd.date_range(end="2005-01-01 10:00:00", freq="H", periods=5)
    dt2 = pd.date_range(start="2005-01-01 13:00:00", freq="H", periods=5)

    dates = dt1.append(dt2).to_series().reset_index(drop=True)

    cases.append(
        {
            "dates": dates,
            "expected_debug_obj": {
                "actual_range_start": dates.loc[0].isoformat(),
                "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
                "message": DataCheckMessageCode.DATETIME_SERIES_IS_NOT_LONG_ENOUGH,
                "estimated_freq": None,
                "estimated_range_start": None,
                "estimated_range_end": None,
                "duplicate_values": [],
                "missing_values": [],
                "extra_values": [],
                "nan_values": [],
            },
        }
    )

    dt1 = pd.date_range(start="2005-01-01 10:00:00", freq="H", periods=50)
    dt2 = pd.date_range(end="2005-01-01 8:00:00", freq="H", periods=50)

    dates = dt1.append(dt2).to_series().reset_index(drop=True)

    cases.append(
        {
            "dates": dates,
            "expected_debug_obj": {
                "actual_range_start": dates.min().isoformat(),
                "actual_range_end": dates.max().isoformat(),
                "message": DataCheckMessageCode.DATETIME_SERIES_IS_NOT_MONOTONIC,
                "estimated_freq": None,
                "estimated_range_start": None,
                "estimated_range_end": None,
                "duplicate_values": [],
                "missing_values": [],
                "extra_values": [],
                "nan_values": [],
            },
        }
    )

    dates = pd.Series(
        [pd.Timestamp(np.nan), pd.Timestamp(np.nan), pd.Timestamp(np.nan)]
    )

    dates = dates.reset_index(drop=True).astype("datetime64[ns]")

    cases.append(
        {
            "dates": dates,
            "expected_debug_obj": {
                "actual_range_start": None,
                "actual_range_end": None,
                "message": DataCheckMessageCode.DATETIME_SERIES_IS_EMPTY,
                "estimated_freq": None,
                "estimated_range_start": None,
                "estimated_range_end": None,
                "duplicate_values": [],
                "missing_values": [],
                "extra_values": [],
                "nan_values": [],
            },
        }
    )

    dt1 = pd.date_range(end="2005-01-01 10:00:00", freq="H", periods=30)
    dt2 = pd.date_range(start="2005-01-01 13:00:00", freq="H", periods=30)

    dates = dt1.append(dt2).to_series().reset_index(drop=True)

    cases.append(
        {
            "dates": dates,
            "expected_debug_obj": {
                "actual_range_start": dates.loc[0].isoformat(),
                "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
                "message": DataCheckMessageCode.DATETIME_SERIES_FREQ_CANNOT_BE_ESTIMATED,
                "estimated_freq": None,
                "estimated_range_start": None,
                "estimated_range_end": None,
                "duplicate_values": [],
                "missing_values": [],
                "extra_values": [],
                "nan_values": [],
            },
        }
    )

    return cases


def case0():
    """
    missing values
    """

    dates = [
        "00:00:00",
        "01:00:00",
        "02:00:00",
        "04:00:00",  # <-- missing index is here
        "05:00:00",
    ]

    dates = [f"2005-01-01T{d}Z" for d in dates]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [],
        "missing_values": [
            {"dt": "2005-01-01T03:00:00", "idx": (HEAD_RANGE_LEN - 1) + 3, "range": 1}
        ],
        "extra_values": [],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case1():
    """
    duplicate values
    """

    dates = [
        "00:00:00",
        "01:00:00",
        "02:00:00",
        "03:00:00",
        "03:00:00",  # <-- duplicate index starts here
        "03:00:00",
        "04:00:00",
        "05:00:00",
    ]
    dates = [f"2005-01-01T{d}Z" for d in dates]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [
            {"dt": "2005-01-01T03:00:00", "idx": (HEAD_RANGE_LEN - 1) + 4, "range": 2}
        ],
        "missing_values": [],
        "extra_values": [],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case2():
    """
    extra values
    """

    dates = [
        "00:00:00",
        "01:00:00",
        "02:00:00",
        "03:00:00",
        "03:10:00",  # <-- extra index is here
        "04:00:00",
        "05:00:00",
    ]
    dates = [f"2005-01-01T{d}Z" for d in dates]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [],
        "missing_values": [],
        "extra_values": [
            {"dt": "2005-01-01T03:10:00", "idx": (HEAD_RANGE_LEN - 1) + 4, "range": 1}
        ],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case3():
    """
    misaligned values - simple
    """

    dates = [
        "00:00:00",
        "01:00:00",
        "02:00:00",
        "03:10:00",  # <-- missing index and extra index is here
        "04:00:00",
        "05:00:00",
    ]
    dates = [f"2005-01-01T{d}Z" for d in dates]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [],
        "missing_values": [
            {"dt": "2005-01-01T03:00:00", "idx": (HEAD_RANGE_LEN - 1) + 3, "range": 1}
        ],
        "extra_values": [
            {"dt": "2005-01-01T03:10:00", "idx": (HEAD_RANGE_LEN - 1) + 3, "range": 1}
        ],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case4():
    """
    misaligned values - complex
    """

    dates = [
        "00:00:00",
        "01:00:00",
        "01:30:00",
        "02:50:00",
        "03:10:00",
        "04:00:00",
        "05:00:00",
    ]
    dates = [f"2005-01-01T{d}Z" for d in dates]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [],
        "missing_values": [
            {"dt": "2005-01-01T02:00:00", "idx": (HEAD_RANGE_LEN - 1) + 2, "range": 2}
        ],
        "extra_values": [
            {"dt": "2005-01-01T01:30:00", "idx": (HEAD_RANGE_LEN - 1) + 2, "range": 3},
        ],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case5():
    """
    bad start
    """

    head_range = pd.DatetimeIndex(["2004-12-31 23:50:00"])
    tail_range = pd.date_range(start="2005-01-01 01:00:00", periods=100, freq="H")[1:]

    dates = head_range.append(tail_range)

    expected_debug_obj = {
        "actual_range_start": dates[0].isoformat(),
        "actual_range_end": dates[-1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": tail_range[0].isoformat(),
        "estimated_range_end": dates[-1].isoformat(),
        "duplicate_values": [],
        "missing_values": [],
        "extra_values": [
            {"dt": "2004-12-31T23:50:00", "idx": 0, "range": 1},
        ],
        "nan_values": [],
    }

    return {"dates": dates.to_series(), "expected_debug_obj": expected_debug_obj}


def case6():
    """
    nan values
    """

    dates = [
        "2005-01-01T00:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T03:00:00.000Z",
        np.nan,
        np.nan,
        "2005-01-01T04:00:00.000Z",
        "2005-01-01T05:00:00.000Z",
    ]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [],
        "missing_values": [],
        "extra_values": [],
        "nan_values": [
            {"dt": None, "idx": (HEAD_RANGE_LEN - 1) + 4, "range": 2},
        ],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case7():
    """
    nans and duplicates
    """

    dates = [
        "2005-01-01T00:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T03:00:00.000Z",
        np.nan,
        np.nan,
        "2005-01-01T04:00:00.000Z",
        "2005-01-01T05:00:00.000Z",
    ]

    dates = pad_datetime_series(dates, freq="H", pad_start=0, pad_end=TAIL_RANGE_LEN)

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [
            {"dt": "2005-01-01T02:00:00", "idx": 3, "range": 2},
        ],
        "missing_values": [],
        "extra_values": [],
        "nan_values": [
            {"dt": None, "idx": 6, "range": 2},
        ],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case8():
    """
    duplicates and extra
    """

    dates = [
        "2005-01-01T00:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T03:00:00.000Z",
        "2005-01-01T03:10:00.000Z",
        "2005-01-01T04:00:00.000Z",
        "2005-01-01T05:00:00.000Z",
    ]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [
            {"dt": "2005-01-01T02:00:00", "idx": (HEAD_RANGE_LEN - 1) + 3, "range": 2},
        ],
        "missing_values": [],
        "extra_values": [
            {"dt": "2005-01-01T03:10:00", "idx": (HEAD_RANGE_LEN - 1) + 6, "range": 1},
        ],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case9():
    """
    duplicates and missing
    """

    dates = [
        "2005-01-01T00:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T03:00:00.000Z",
        "2005-01-01T05:00:00.000Z",
    ]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [
            {"dt": "2005-01-01T02:00:00", "idx": (HEAD_RANGE_LEN - 1) + 3, "range": 2},
        ],
        "missing_values": [
            {"dt": "2005-01-01T04:00:00", "idx": (HEAD_RANGE_LEN - 1) + 4, "range": 1},
        ],
        "extra_values": [],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case10():
    dates = [
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T03:00:00.000Z",
        "2005-01-01T05:00:00.000Z",
        "2005-01-01T06:00:00.000Z",
        "2005-01-01T06:20:00.000Z",
        "2005-01-01T07:00:00.000Z",
        "2005-01-01T08:00:00.000Z",
    ]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [
            {"dt": "2005-01-01T01:00:00", "idx": (HEAD_RANGE_LEN - 1) + 1, "range": 2},
        ],
        "missing_values": [
            {"dt": "2005-01-01T04:00:00", "idx": (HEAD_RANGE_LEN - 1) + 3, "range": 1},
        ],
        "extra_values": [
            {"dt": "2005-01-01T06:20:00", "idx": (HEAD_RANGE_LEN - 1) + 7, "range": 1},
        ],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


def case11():
    """
    Multiple Ranges
    """
    dates = [
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T01:00:00.000Z",
        "2005-01-01T02:00:00.000Z",
        "2005-01-01T03:00:00.000Z",
        "2005-01-01T04:00:00.000Z",
        "2005-01-01T05:00:00.000Z",
        "2005-01-01T06:00:00.000Z",
        "2005-01-01T06:00:00.000Z",
        "2005-01-01T06:00:00.000Z",
        "2005-01-01T06:00:00.000Z",
        "2005-01-01T07:00:00.000Z",
        "2005-01-01T08:00:00.000Z",
    ]

    dates = pad_datetime_series(
        dates, freq="H", pad_start=HEAD_RANGE_LEN, pad_end=TAIL_RANGE_LEN
    )

    expected_debug_obj = {
        "actual_range_start": dates.loc[0].isoformat(),
        "actual_range_end": dates.loc[len(dates) - 1].isoformat(),
        "message": None,
        "estimated_freq": "H",
        "estimated_range_start": dates.loc[0].isoformat(),
        "estimated_range_end": dates.loc[len(dates) - 1].isoformat(),
        "duplicate_values": [
            {"dt": "2005-01-01T01:00:00", "idx": (HEAD_RANGE_LEN - 1) + 1, "range": 2},
            {"dt": "2005-01-01T06:00:00", "idx": (HEAD_RANGE_LEN - 1) + 8, "range": 3},
        ],
        "missing_values": [],
        "extra_values": [],
        "nan_values": [],
    }

    return {"dates": dates, "expected_debug_obj": expected_debug_obj}


inferrable_freq_fixtures = [
    case0(),
    case1(),
    case2(),
    case3(),
    case4(),
    case5(),
    case6(),
    case7(),
    case8(),
    case9(),
    case10(),
    case11(),
]



from collections import defaultdict
from timeit import default_timer as timer

import numpy as np

from ._bin_numeric_cols_into_categories import _bin_numeric_cols_into_categories
from ._calculate_dependence_measure import _calculate_dependence_measure
from ._calculate_max_dependence_for_pair import _calculate_max_dependence_for_pair
from ._cast_nullable_int_and_datetime_to_int import (
    _cast_nullable_int_and_datetime_to_int,
)
from ._parse_measures import _parse_measures

from woodwork.accessor_utils import _is_dask_dataframe, _is_spark_dataframe
from woodwork.utils import CallbackCaller, get_valid_mi_types, get_valid_pearson_types


def _get_dependence_dict(
    dataframe,
    measures,
    num_bins=10,
    nrows=None,
    include_index=False,
    callback=None,
    extra_stats=False,
    min_shared=25,
    random_seed=0,
):
    """Calculates dependence measures between all pairs of columns in the DataFrame that
    support measuring dependence. Supports boolean, categorical, datetime, and numeric data.
    Call woodwork.utils.get_valid_mi_types and woodwork.utils.get_valid_pearson_types
    for complete lists of supported Logical Types.

    Args:
        dataframe (pd.DataFrame): Data containing Woodwork typing information
            from which to calculate dependence.
        measures (list or str): Which dependence measures to calculate.
            A list of measures can be provided to calculate multiple
            measures at once.  Valid measure strings:

                - "pearson": calculates the Pearson correlation coefficient
                - "mutual_info": calculates the mutual information between columns
                - "max":  max(abs(pearson), mutual) for each pair of columns
                - "all": includes columns for "pearson", "mutual_info", and "max"
        num_bins (int): Determines number of bins to use for converting numeric
            features into categorical.  Default to 10. Pearson calculation does
            not use binning.
        nrows (int): The number of rows to sample for when determining dependence.
            If specified, samples the desired number of rows from the data.
            Defaults to using all rows.
        include_index (bool): If True, the column specified as the index will be
            included as long as its LogicalType is valid for measuring dependence.
            If False, the index column will not be considered. Defaults to False.
        callback (callable, optional): function to be called with incremental updates. Has the following parameters:

            - update (int): change in progress since last call
            - progress (int): the progress so far in the calculations
            - total (int): the total number of calculations to do
            - unit (str): unit of measurement for progress/total
            - time_elapsed (float): total time in seconds elapsed since start of call
        extra_stats (bool):  If True, additional column "shared_rows"
            recording the number of shared non-null rows for a column
            pair will be included with the dataframe.  If the "max"
            measure is being used, a "measure_used" column will be added
            that records whether Pearson or mutual information was the
            maximum dependence for a particular row. Defaults to False.
        min_shared (int): The number of shared non-null rows needed to
            calculate.  Less rows than this will be considered too sparse
            to measure accurately and will return a NaN value. Must be
            non-negative. Defaults to 25.
        random_seed (int): Seed for the random number generator. Defaults to 0.
    Returns:
        list(dict): A list containing dictionaries that have keys `column_1`,
        `column_2`, and keys for the specified dependence measures. The list is
        sorted in decending order by the first specified measure.
        Dependence information values are between 0 (no dependence) and 1
        (perfect dependency). For Pearson, values range from -1 to 1 but 0 is
        still no dependence.
    """
    start_time = timer()

    returned_measures, calc_order, calc_max = _parse_measures(measures)

    unit = "calculations"

    # get valid columns for dependence calculations
    if "pearson" in calc_order:
        pearson_types = get_valid_pearson_types()
        pearson_columns = _get_valid_columns(dataframe, pearson_types)
        valid_columns = pearson_columns
    if "mutual_info" in calc_order:
        mi_types = get_valid_mi_types()
        mutual_columns = _get_valid_columns(dataframe, mi_types)
        # pearson columns are a subset of mutual columns
        valid_columns = mutual_columns

    index = dataframe.ww.index
    if not include_index and index is not None and index in valid_columns:
        valid_columns.remove(index)

    data = dataframe.loc[:, valid_columns]
    # cut off data if necessary
    if _is_dask_dataframe(data):
        data = data.compute()
    elif _is_spark_dataframe(dataframe):
        data = data.to_pandas()
    if nrows is not None and nrows < data.shape[0]:
        data = data.sample(nrows, random_state=random_seed)

    notna_mask = data.notnull()
    not_null_cols = data.columns[notna_mask.any()]
    not_null_col_set = set(not_null_cols)
    if not_null_col_set != set(valid_columns):
        data = data.loc[:, not_null_cols]

    p = 0  # number of pearson columns
    m = 0  # number of mutual columns
    if "pearson" in calc_order:
        pearson_columns = [col for col in pearson_columns if col in not_null_col_set]
        p = len(pearson_columns)
    if "mutual_info" in calc_order:
        mutual_columns = [col for col in mutual_columns if col in not_null_col_set]
        m = len(mutual_columns)
    n = max(m, p)

    # combinations in a loop is n! / 2 / (n - 2)! which reduces to (n) (n - 1) / 2
    def _num_calc_steps(n):
        return (n * n - n) / 2

    # Assume 1 unit for preprocessing, n for handling nulls, m for binning numerics
    total_loops = 1 + n + m + _num_calc_steps(p) + _num_calc_steps(m)
    callback_caller = CallbackCaller(callback, unit, total_loops, start_time=start_time)
    callback_caller.update(1)

    # split dataframe into dict of series so we can drop nulls on a per-column basis
    data = {col: data[col].dropna() for col in data}

    # cast nullable type to non-nullable (needed for both pearson and mutual)
    _cast_nullable_int_and_datetime_to_int(data, dataframe.ww.columns)
    callback_caller.update(n)

    results = defaultdict(dict)

    for measure in calc_order:
        if measure == "mutual_info":
            _bin_numeric_cols_into_categories(dataframe.ww.schema, data, num_bins)
            callback_caller.update(n)
            col_names = mutual_columns
        elif measure == "pearson":
            col_names = pearson_columns

        _calculate_dependence_measure(
            measure=measure,
            data=data,
            results=results,
            callback_caller=callback_caller,
            notna_mask=notna_mask,
            min_shared=min_shared,
            col_names=col_names,
        )

    for result in results.values():
        if calc_max:
            _calculate_max_dependence_for_pair(
                result=result,
                min_shared=min_shared,
                extra_stats=extra_stats,
            )
            if returned_measures == ["max"]:
                # remove measurements not expected in returned dictionary
                del result["mutual_info"]
                if "pearson" in result:
                    del result["pearson"]

        # Remove cached info not expected in result by user
        if "num_union" in result:
            del result["num_union"]
        if not extra_stats:
            del result["shared_rows"]

    results = list(results.values())

    def sort_key(result):
        key = abs(result[returned_measures[0]])
        if np.isnan(key):
            key = -1
        return key

    results.sort(key=sort_key, reverse=True)

    return results


def _get_valid_columns(dataframe, valid_types):
    valid_columns = [
        col_name
        for col_name, col in dataframe.ww.columns.items()
        if type(col.logical_type) in valid_types
    ]
    return valid_columns



from .parquet_deserializer import ParquetDeserializer
from .csv_deserializer import CSVDeserializer
from .feather_deserializer import FeatherDeserializer
from .arrow_deserializer import ArrowDeserializer
from .orc_deserializer import OrcDeserializer
from .pickle_deserializer import PickleDeserializer

from .utils import _get_deserializer



import re
import warnings

import pandas as pd
from pandas.api import types as pdtypes

from woodwork.accessor_utils import _is_dask_series, _is_spark_series
from woodwork.config import config
from woodwork.exceptions import (
    TypeConversionError,
    TypeConversionWarning,
    TypeValidationError,
)
from woodwork.type_sys.utils import _get_specified_ltype_params
from woodwork.utils import (
    _infer_datetime_format,
    _is_valid_latlong_series,
    _reformat_to_latlong,
    camel_to_snake,
    import_or_none,
)

dd = import_or_none("dask.dataframe")
ps = import_or_none("pyspark.pandas")


class ClassNameDescriptor(object):
    """Descriptor to convert a class's name from camelcase to snakecase"""

    def __get__(self, instance, class_):
        return camel_to_snake(class_.__name__)


class LogicalTypeMetaClass(type):
    def __repr__(cls):
        return cls.__name__


class LogicalType(object, metaclass=LogicalTypeMetaClass):
    """Base class for all other Logical Types"""

    type_string = ClassNameDescriptor()
    primary_dtype = "string"
    backup_dtype = None
    standard_tags = set()

    def __eq__(self, other, deep=False):
        return isinstance(other, self.__class__) and _get_specified_ltype_params(
            other
        ) == _get_specified_ltype_params(self)

    def __str__(self):
        return str(self.__class__)

    @classmethod
    def _get_valid_dtype(cls, series_type):
        """Return the dtype that is considered valid for a series with the given logical_type"""
        if ps and series_type == ps.Series and cls.backup_dtype:
            return cls.backup_dtype
        else:
            return cls.primary_dtype

    def transform(self, series):
        """Converts the series dtype to match the logical type's if it is different."""
        new_dtype = self._get_valid_dtype(type(series))
        if new_dtype != str(series.dtype):
            # Update the underlying series
            try:
                series = series.astype(new_dtype)
            except (TypeError, ValueError):
                raise TypeConversionError(series, new_dtype, type(self))
        return series

    def validate(self, series, *args, **kwargs):
        """Validates that a logical type is consistent with the series dtype. Performs additional type
        specific validation, as required. When the series' dtype does not match the logical types' required dtype,
        raises a TypeValidationError."""
        valid_dtype = self._get_valid_dtype(type(series))
        if valid_dtype != str(series.dtype):
            raise TypeValidationError(
                f"Series dtype '{series.dtype}' is incompatible with {self.type_string} dtype."
            )


class Address(LogicalType):
    """Represents Logical Types that contain address values.

    Examples:
        .. code-block:: python

            ['1 Miller Drive, New York, NY 12345', '1 Berkeley Street, Boston, MA 67891']
            ['26387 Russell Hill, Dallas, TX 34521', '54305 Oxford Street, Seattle, WA 95132']
    """

    primary_dtype = "string"


class Age(LogicalType):
    """Represents Logical Types that contain whole numbers indicating a person's age.
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [15, 22, 45]
            [30, 62, 87]
    """

    primary_dtype = "int64"
    standard_tags = {"numeric"}

    def validate(self, series, return_invalid_values=True):
        """Validates age values by checking for non-negative values.

        Args:
            series (Series): Series of age values
            return_invalid_values (bool): Whether or not to return invalid age values

        Returns:
            Series: If return_invalid_values is True, returns invalid age values.
        """
        return _validate_not_negative(series, return_invalid_values)


class AgeFractional(LogicalType):
    """Represents Logical Types that contain non-negative floating point numbers indicating a person's age.
    Has 'numeric' as a standard tag. May also contain null values.

    Examples:
        .. code-block:: python

            [0.34, 24.34, 45.0]
            [30.5, 62.82, np.nan]
    """

    primary_dtype = "float64"
    standard_tags = {"numeric"}

    def validate(self, series, return_invalid_values=True):
        """Validates age values by checking for non-negative values.

        Args:
            series (Series): Series of age values
            return_invalid_values (bool): Whether or not to return invalid age values

        Returns:
            Series: If return_invalid_values is True, returns invalid age values.
        """
        return _validate_not_negative(series, return_invalid_values)


class AgeNullable(LogicalType):
    """Represents Logical Types that contain whole numbers indicating a person's age.
    Has 'numeric' as a standard tag. May also contain null values.

    Examples:
        .. code-block:: python

            [np.nan, 22, 45]
            [30, 62, np.nan]
    """

    primary_dtype = "Int64"
    standard_tags = {"numeric"}

    def validate(self, series, return_invalid_values=True):
        """Validates age values by checking for non-negative values.

        Args:
            series (Series): Series of age values
            return_invalid_values (bool): Whether or not to return invalid age values

        Returns:
            Series: If return_invalid_values is True, returns invalid age values.
        """
        return _validate_not_negative(series, return_invalid_values)


class Boolean(LogicalType):
    """Represents Logical Types that contain binary values indicating true/false.

    Examples:
        .. code-block:: python

            [True, False, True]
            [0, 1, 1]
    """

    primary_dtype = "bool"


class BooleanNullable(LogicalType):
    """Represents Logical Types that contain binary values indicating true/false.
    May also contain null values.

    Examples:
        .. code-block:: python

            [True, False, None]
            [0, 1, 1]
    """

    primary_dtype = "boolean"


class Categorical(LogicalType):
    """Represents Logical Types that contain unordered discrete values that fall
    into one of a set of possible values. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["red", "green", "blue"]
            ["produce", "dairy", "bakery"]
            [3, 1, 2]
    """

    primary_dtype = "category"
    backup_dtype = "string"
    standard_tags = {"category"}

    def __init__(self, encoding=None):
        # encoding dict(str -> int)
        # user can specify the encoding to use downstream
        pass


class CountryCode(LogicalType):
    """Represents Logical Types that use the ISO-3166 standard country code to represent countries.
    ISO 3166-1 (countries) are supported. These codes should be in the Alpha-2 format.

    Examples:
        .. code-block:: python

            ["AU", "US", "UA"]
            ["GB", "NZ", "DE"]
    """

    primary_dtype = "category"
    backup_dtype = "string"
    standard_tags = {"category"}


class CurrencyCode(LogicalType):
    """Represents Logical Types that use the ISO-4217 internation standard currency code to represent currencies.

    Examples:
        .. code-block:: python

            ["GBP", "JPY", "USD"]
            ["SAR", "EUR", "CZK"]
    """

    primary_dtype = "category"
    backup_dtype = "string"
    standard_tags = {"category"}


class Datetime(LogicalType):
    """Represents Logical Types that contain date and time information.

    Args:
        datetime_format (str): Desired datetime format for data

    Examples:
        .. code-block:: python

            ["2020-09-10",
             "2020-01-10 00:00:00",
             "01/01/2000 08:30"]
    """

    primary_dtype = "datetime64[ns]"
    datetime_format = None

    def __init__(self, datetime_format=None, timezone=None):
        self.datetime_format = datetime_format
        self.timezone = timezone

    def _remove_timezone(self, series):
        """Removes timezone from series and stores in logical type."""
        if hasattr(series.dtype, "tz") and series.dtype.tz:
            self.timezone = str(series.dtype.tz)
            series = series.dt.tz_localize(None)
        return series

    def transform(self, series):
        """Converts the series data to a formatted datetime. Datetime format will be inferred if datetime_format is None."""
        new_dtype = self._get_valid_dtype(type(series))
        series = self._remove_timezone(series)
        series_dtype = str(series.dtype)

        if new_dtype != series_dtype:
            self.datetime_format = self.datetime_format or _infer_datetime_format(
                series
            )
            utc = self.datetime_format and self.datetime_format.endswith("%z")
            if _is_dask_series(series):
                name = series.name
                series = dd.to_datetime(
                    series, format=self.datetime_format, errors="coerce", utc=utc
                )
                series.name = name
            elif _is_spark_series(series):
                series = ps.Series(
                    ps.to_datetime(
                        series.to_numpy(), format=self.datetime_format, errors="coerce"
                    ),
                    name=series.name,
                )
            else:
                try:
                    series = pd.to_datetime(
                        series, format=self.datetime_format, utc=utc
                    )
                except (TypeError, ValueError):
                    warnings.warn(
                        f"Some rows in series '{series.name}' are incompatible with datetime format "
                        f"'{self.datetime_format}' and have been replaced with null values. You may be "
                        "able to fix this by using an instantiated Datetime logical type with a different format "
                        "string specified for this column during Woodwork initialization.",
                        TypeConversionWarning,
                    )
                    series = pd.to_datetime(
                        series,
                        format=self.datetime_format,
                        errors="coerce",
                        utc=utc,
                    )

        series = self._remove_timezone(series)
        return super().transform(series)


class Double(LogicalType):
    """Represents Logical Types that contain positive and negative numbers, some of
    which include a fractional component. Includes zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [1.2, 100.4, 3.5]
            [-15.34, 100, 58.3]
    """

    primary_dtype = "float64"
    standard_tags = {"numeric"}


class Integer(LogicalType):
    """Represents Logical Types that contain positive and negative numbers
    without a fractional component, including zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [100, 35, 0]
            [-54, 73, 11]
    """

    primary_dtype = "int64"
    standard_tags = {"numeric"}


class IntegerNullable(LogicalType):
    """Represents Logical Types that contain positive and negative numbers
    without a fractional component, including zero (0). May contain null
    values. Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [100, 35, np.nan]
            [-54, 73, 11]
    """

    primary_dtype = "Int64"
    standard_tags = {"numeric"}


class EmailAddress(LogicalType):
    """Represents Logical Types that contain email address values.

    Examples:
        .. code-block:: python

            ["john.smith@example.com",
             "support@example.com",
             "team@example.com"]
    """

    primary_dtype = "string"

    def validate(self, series, return_invalid_values=False):
        """Validates email address values based on the regex in the config.

        Args:
            series (Series): Series of email address values
            return_invalid_values (bool): Whether or not to return invalid email address values

        Returns:
            Series: If return_invalid_values is True, returns invalid email address.
        """
        return _regex_validate("email_inference_regex", series, return_invalid_values)


class Filepath(LogicalType):
    """Represents Logical Types that specify locations of directories and files
    in a file system.

    Examples:
        .. code-block:: python

            ["/usr/local/bin",
             "/Users/john.smith/dev/index.html",
             "/tmp"]
    """

    primary_dtype = "string"


class PersonFullName(LogicalType):
    """Represents Logical Types that may contain first, middle and last names,
    including honorifics and suffixes.

    Examples:
        .. code-block:: python

            ["Mr. John Doe, Jr.",
             "Doe, Mrs. Jane",
             "James Brown"]
    """

    primary_dtype = "string"


class IPAddress(LogicalType):
    """Represents Logical Types that contain IP addresses, including both
    IPv4 and IPv6 addresses.

    Examples:
        .. code-block:: python

            ["172.16.254.1",
             "192.0.0.0",
             "2001:0db8:0000:0000:0000:ff00:0042:8329"]
    """

    primary_dtype = "string"


class LatLong(LogicalType):
    """Represents Logical Types that contain latitude and longitude values in decimal degrees.

    Note:
        LatLong values will be stored with the object dtype as a
        tuple of floats (or a list of floats for Spark DataFrames)
        and must contain only two values.

        Null latitude or longitude values will be stored as np.nan, and
        a fully null LatLong (np.nan, np.nan) will be stored as just a
        single nan.

    Examples:
        .. code-block:: python

            [(33.670914, -117.841501),
             (40.423599, -86.921162),
             (-45.031705, nan)]
    """

    primary_dtype = "object"

    def transform(self, series):
        """Formats a series to be a tuple (or list for Spark) of two floats."""
        if _is_dask_series(series):
            name = series.name
            meta = (series, tuple([float, float]))
            series = series.apply(_reformat_to_latlong, meta=meta)
            series.name = name
        elif _is_spark_series(series):
            formatted_series = series.to_pandas().apply(
                _reformat_to_latlong, is_spark=True
            )
            series = ps.from_pandas(formatted_series)
        else:
            series = series.apply(_reformat_to_latlong)

        return super().transform(series)

    def validate(self, series, return_invalid_values=False):
        # TODO: we'll want to actually handle return_invalid_values in the ordinal and latlong logical types.
        super().validate(series)
        if not _is_valid_latlong_series(series):
            raise TypeValidationError(
                "Cannot initialize Woodwork. Series does not contain properly formatted "
                "LatLong data. Try reformatting before initializing or use the "
                "woodwork.init_series function to initialize."
            )


class NaturalLanguage(LogicalType):
    """Represents Logical Types that contain text or characters representing
    natural human language

    Examples:
        .. code-block:: python

            ["This is a short sentence.",
             "I like to eat pizza!",
             "When will humans go to mars?"]
    """

    primary_dtype = "string"


class Unknown(LogicalType):
    """Represents Logical Types that cannot be inferred as a specific Logical Type. It is assumed to contain string data.

    Examples:
        .. code-block:: python

            ["ax23n9ck23l",
             "1,28&*_%*&&xejc",
             "xnmvz@@Dcmeods-0"]

    """

    primary_dtype = "string"


class Ordinal(LogicalType):
    """Represents Logical Types that contain ordered discrete values.
    Has 'category' as a standard tag.

    Args:
        order (list or tuple): An list or tuple specifying the order of the ordinal
            values from low to high. The underlying series cannot contain values that
            are not present in the order values.

    Examples:
        .. code-block:: python

            ["first", "second", "third"]
            ["bronze", "silver", "gold"]
    """

    primary_dtype = "category"
    backup_dtype = "string"
    standard_tags = {"category"}

    def __init__(self, order=None):
        self.order = order

    def _validate_order_values(self, series):
        """Make sure order values are properly defined and confirm the supplied series
        does not contain any values that are not in the specified order values"""
        if self.order is None:
            raise TypeError("Must use an Ordinal instance with order values defined")
        elif not isinstance(self.order, (list, tuple)):
            raise TypeError("Order values must be specified in a list or tuple")
        if len(self.order) != len(set(self.order)):
            raise ValueError("Order values cannot contain duplicates")

        if isinstance(series, pd.Series):
            missing_order_vals = set(series.dropna().values).difference(self.order)
            if missing_order_vals:
                error_msg = (
                    f"Ordinal column {series.name} contains values that are not present "
                    f"in the order values provided: {sorted(list(missing_order_vals))}"
                )
                raise ValueError(error_msg)

    def transform(self, series):
        """Validates the series and converts the dtype to match the logical type's if it is different."""
        self._validate_order_values(series)

        typed_ser = super().transform(series)
        if pdtypes.is_categorical_dtype(typed_ser.dtype):
            typed_ser = typed_ser.cat.set_categories(self.order, ordered=True)

        return typed_ser

    def validate(self, series, return_invalid_values=False):
        # TODO: we'll want to actually handle return_invalid_values in the ordinal and latlong logical types.
        super().validate(series)
        self._validate_order_values(series)

    def __str__(self):
        return "{}: {}".format(self.__class__, self.order)


class PhoneNumber(LogicalType):
    """Represents Logical Types that contain numeric digits and characters
    representing a phone number.

    Examples:
        .. code-block:: python

            ["1-(555)-123-5495",
             "+1-555-123-5495",
             "5551235495"]
    """

    primary_dtype = "string"

    def validate(self, series, return_invalid_values=False):
        """Validates PhoneNumber values based on the regex in the config.
        By default, this validates US/Canada-based phone numbers.

        Args:
            series (Series): Series of phone number values.
            return_invalid_values (bool): Whether or not to return invalid phone numbers.

        Returns:
            Series: If return_invalid_values is True, returns invalid phone numbers.
        """
        return _regex_validate("phone_inference_regex", series, return_invalid_values)


class SubRegionCode(LogicalType):
    """Represents Logical Types that use the ISO-3166 standard sub-region code to
    represent a portion of a larger geographic region. ISO 3166-2 (sub-regions)
    codes are supported. These codes should be in the Alpha-2 format.

    Examples:
        .. code-block:: python

            ["US-CO", "US-MA", "US-CA"]
            ["AU-NSW", "AU-TAS", "AU-QLD"]
    """

    primary_dtype = "category"
    backup_dtype = "string"
    standard_tags = {"category"}


class Timedelta(LogicalType):
    """Represents Logical Types that contain values specifying a duration of time

    Examples:
        .. code-block:: python

            [pd.Timedelta('1 days 00:00:00'),
             pd.Timedelta('-1 days +23:40:00'),
             pd.Timedelta('4 days 12:00:00')]
    """

    primary_dtype = "timedelta64[ns]"


class URL(LogicalType):
    """Represents Logical Types that contain URLs, which may include protocol, hostname
    and file name

    Examples:
        .. code-block:: python

            ["http://google.com",
             "https://example.com/index.html",
             "example.com"]
    """

    primary_dtype = "string"

    def validate(self, series, return_invalid_values=False):
        """Validates URL values based on the regex in the config.

        Args:
            series (Series): Series of URL values
            return_invalid_values (bool): Whether or not to return invalid URLs

        Returns:
            Series: If return_invalid_values is True, returns invalid URLs.
        """
        return _regex_validate("url_inference_regex", series, return_invalid_values)


class PostalCode(LogicalType):
    """Represents Logical Types that contain a series of postal codes for
    representing a group of addresses. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["90210"
             "60018-0123",
             "10010"]
    """

    primary_dtype = "category"
    backup_dtype = "string"
    standard_tags = {"category"}

    def validate(self, series, return_invalid_values=False):
        """Validates PostalCode values based on the regex in the config. Currently only validates US Postal codes.

        Args:
            series (Series): Series of PostalCode values.
            return_invalid_values (bool): Whether or not to return invalid PostalCodes.

        Returns:
            Series: If return_invalid_values is True, returns invalid PostalCodes.
        """
        return _regex_validate(
            "postal_code_inference_regex", series, return_invalid_values
        )


_NULLABLE_PHYSICAL_TYPES = {
    "boolean",
    "category",
    "datetime64[ns]",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Float32",
    "Float64",
    "float16",
    "float32",
    "float64",
    "float128",
    "object",
    "string",
    "timedelta64[ns]",
}


def _regex_validate(regex_key, series, return_invalid_values):
    """Validates data values based on the logical type regex in the config."""
    regex = config.get_option(regex_key)

    if _is_spark_series(series):

        def match(x):
            if isinstance(x, str):
                return bool(re.match(regex, x))

        invalid = series.apply(match).astype("boolean") == False  # noqa: E712

    else:
        invalid = ~series.str.match(regex).astype("boolean")

    if return_invalid_values:
        return series[invalid]

    else:
        any_invalid = invalid.any()
        if dd and isinstance(any_invalid, dd.core.Scalar):
            any_invalid = any_invalid.compute()

        if any_invalid:
            type_string = {
                "url_inference_regex": "url",
                "email_inference_regex": "email address",
                "phone_inference_regex": "phone number",
                "postal_code_inference_regex": "postal code",
            }[regex_key]

            info = f"Series {series.name} contains invalid {type_string} values. "
            info += f"The {regex_key} can be changed in the config if needed."
            raise TypeValidationError(info)


def _validate_not_negative(series, return_invalid_values):
    """Validates data values are non-negative."""
    invalid = series.lt(0)
    if return_invalid_values:
        return series[invalid]

    else:
        any_invalid = invalid.any()
        if dd and isinstance(any_invalid, dd.core.Scalar):
            any_invalid = any_invalid.compute()

        if any_invalid:
            info = f"Series {series.name} contains negative values."
            raise TypeValidationError(info)



from ._constants import FREQ_INFERENCE_THRESHOLD, NON_INFERABLE_FREQ


def _determine_most_likely_freq(alias_dict, threshold=FREQ_INFERENCE_THRESHOLD):
    """Determine most likely frequency, given the input alias_dict

    Args:
        alias_dict (dict): A dictionary where the key values are a Pandas alias string or a string representing that
            a frequency cannot be inferred. The value of the dictionary is the following:

            - alias (str): The Pandas Freq Alias
            - count (int): The number of windows where this alias is valid
            - min_dt (pd.TimeStamp): The earliest timestamp for this frequency.
            - max_dt (pd.TimeStamp): The latest timestamp for this frequency.

    Returns:
        (list(RangeObject)): A list of RangeObject data objects. A RangeObject has the following properties:

        - dt: an ISO 8601 formatted string of the first NaN timestamp
        - idx: first index of the NaN timestamp. Index is relative to estimated timeseries
        - range: the number of sequential elements that are NaN
    """
    alias_info = alias_dict.values()

    n_total = sum([x["count"] for x in alias_info])
    sorted_aliases = sorted(alias_info, key=lambda item: item["count"], reverse=True)

    most_likely_alias = sorted_aliases[0]["alias"]
    most_likely_count = sorted_aliases[0]["count"]

    if most_likely_count / n_total < threshold:
        return None

    if most_likely_alias is not NON_INFERABLE_FREQ:
        return most_likely_alias
    else:
        return None



import warnings

from woodwork.exceptions import ParametersIgnoredWarning


def _parse_measures(measures):
    """Validates the provided measures argument and returns the list of measures
    included in the final dataframe/dictionary, an ordered list of measures to
    calculate, and whether to calculate the max score across measures."""
    if not isinstance(measures, list):
        if not isinstance(measures, str):
            raise TypeError(f"Supplied measure {measures} is not a string")
        measures = [measures]

    if len(measures) == 0:
        raise ValueError("No measures supplied")

    calc_pearson = False
    calc_mutual = False
    calc_max = False
    calc_order = []
    for measure in measures:
        if measure == "all":
            if not measures == ["all"]:
                warnings.warn(
                    ParametersIgnoredWarning(
                        "additional measures to 'all' measure found; 'all' should be used alone"
                    )
                )
            measures = ["max", "pearson", "mutual_info"]
            calc_pearson = True
            calc_mutual = True
            calc_max = True
            break
        elif measure == "pearson":
            calc_pearson = True
        elif measure == "mutual_info":
            calc_mutual = True
        elif measure == "max":
            calc_pearson = True
            calc_mutual = True
            calc_max = True
        else:
            raise ValueError("Unrecognized dependence measure %s" % measure)

    if calc_pearson:
        calc_order.append("pearson")
    if calc_mutual:
        calc_order.append("mutual_info")

    return measures, calc_order, calc_max



from woodwork.serializers import (
    ArrowSerializer,
    CSVSerializer,
    FeatherSerializer,
    OrcSerializer,
    ParquetSerializer,
    PickleSerializer,
)

# Dictionary used to get the corret serializer from a given format
FORMAT_TO_SERIALIZER = {
    CSVSerializer.format: CSVSerializer,
    PickleSerializer.format: PickleSerializer,
    ParquetSerializer.format: ParquetSerializer,
    ArrowSerializer.format: ArrowSerializer,
    FeatherSerializer.format: FeatherSerializer,
    OrcSerializer.format: OrcSerializer,
}


def get_serializer(format=None):
    """Get serializer class based on format"""
    serializer = FORMAT_TO_SERIALIZER.get(format)
    if serializer is None:
        error = "must be one of the following formats: {}"
        raise ValueError(error.format(", ".join(FORMAT_TO_SERIALIZER.keys())))

    return serializer



import re

import pandas as pd
import pytest

from woodwork.accessor_utils import _is_spark_series, init_series
from woodwork.exceptions import TypeConversionWarning, TypeValidationError
from woodwork.logical_types import (
    URL,
    Age,
    AgeFractional,
    AgeNullable,
    Boolean,
    Categorical,
    Datetime,
    EmailAddress,
    LatLong,
    Ordinal,
    PhoneNumber,
    PostalCode,
)
from woodwork.tests.testing_utils.table_utils import to_pandas
from woodwork.utils import import_or_none

ps = import_or_none("pyspark.pandas")


def test_logical_eq():
    assert Boolean == Boolean
    assert Boolean() == Boolean()
    assert Categorical != Boolean
    assert Datetime != Datetime()
    assert Datetime() == Datetime(datetime_format=None)
    assert Datetime() != Datetime(datetime_format="%Y-%m-%d")


def test_logical_repr():
    assert isinstance(repr(Boolean), str)
    assert repr(Boolean) == "Boolean"
    assert isinstance(repr(Categorical), str)
    assert repr(Categorical) == "Categorical"


def test_instantiated_type_str():
    assert str(Categorical()) == "Categorical"
    assert str(Boolean()) == "Boolean"


def test_ordinal_order_errors():
    series = pd.Series([1, 2, 3]).astype("category")

    with pytest.raises(
        TypeError, match="Order values must be specified in a list or tuple"
    ):
        Ordinal(order="not_valid").transform(series)

    with pytest.raises(ValueError, match="Order values cannot contain duplicates"):
        Ordinal(order=["a", "b", "b"]).transform(series)


def test_ordinal_init_with_order():
    order = ["bronze", "silver", "gold"]
    ordinal_from_list = Ordinal(order=order)
    assert ordinal_from_list.order == order
    assert str(ordinal_from_list) == "Ordinal: ['bronze', 'silver', 'gold']"

    order = ("bronze", "silver", "gold")
    ordinal_from_tuple = Ordinal(order=order)
    assert ordinal_from_tuple.order == order
    assert str(ordinal_from_list) == "Ordinal: ['bronze', 'silver', 'gold']"


def test_ordinal_transform_validates(ordinal_transform_series_pandas) -> None:
    typ = Ordinal(order=None)
    with pytest.raises(TypeError, match=r"order values defined"):
        typ.transform(ordinal_transform_series_pandas)


def test_ordinal_transform_pandas(ordinal_transform_series_pandas) -> None:
    order = [2, 1, 3]
    typ = Ordinal(order=order)
    ser_ = typ.transform(ordinal_transform_series_pandas)

    assert ser_.dtype == "category"
    pd.testing.assert_index_equal(ser_.cat.categories, pd.Int64Index(order))


def test_ordinal_transform_dask(ordinal_transform_series_dask) -> None:
    order = [2, 1, 3]
    typ = Ordinal(order=order)
    ser_ = typ.transform(ordinal_transform_series_dask).compute()

    assert ser_.dtype == "category"
    pd.testing.assert_index_equal(ser_.cat.categories, pd.Int64Index(order))


def test_ordinal_transform_spark(ordinal_transform_series_spark) -> None:
    order = [2, 1, 3]
    typ = Ordinal(order=order)
    ser_ = typ.transform(ordinal_transform_series_spark)

    assert ser_.dtype == pd.StringDtype()


def test_get_valid_dtype(sample_series):
    valid_dtype = Categorical._get_valid_dtype(type(sample_series))
    if _is_spark_series(sample_series):
        assert valid_dtype == "string"
    else:
        assert valid_dtype == "category"

    valid_dtype = Boolean._get_valid_dtype(type(sample_series))
    assert valid_dtype == "bool"


def test_latlong_transform(latlong_df):
    df_type = str(type(latlong_df))
    dask = "dask" in df_type
    spark = "spark" in df_type
    nan = float("nan")

    expected_data = {
        "tuple_ints": [(1.0, 2.0), (3.0, 4.0)],
        "tuple_strings": [(1.0, 2.0), (3.0, 4.0)],
        "string_tuple": [(1.0, 2.0), (3.0, 4.0)],
        "bracketless_string_tuple": [(1.0, 2.0), (3.0, 4.0)],
        "list_strings": [(1.0, 2.0), (3.0, 4.0)],
        "combo_tuple_types": [(1.0, 2.0), (3.0, 4.0)],
        "null_value": [nan, (3.0, 4.0)],
        "null_latitude": [(nan, 2.0), (3.0, 4.0)],
        "both_null": [(nan, nan), (3.0, 4.0)],
    }

    latlong = LatLong()
    for column in latlong_df:
        series = latlong_df[column]
        actual = latlong.transform(series)

        if dask:
            actual = actual.compute()
        elif spark:
            actual = actual.to_pandas()

        expected = pd.Series(expected_data[column], name=column)
        pd.testing.assert_series_equal(actual, expected)


def test_latlong_validate(latlong_df):
    error_message = re.escape(
        "Cannot initialize Woodwork. Series does not contain properly formatted "
        "LatLong data. Try reformatting before initializing or use the "
        "woodwork.init_series function to initialize."
    )
    latlong = LatLong()
    series = latlong_df["tuple_ints"]
    new_series = init_series(series, logical_type=LatLong)
    latlong.validate(new_series)
    with pytest.raises(TypeValidationError, match=error_message):
        latlong.validate(series)


def test_datetime_transform(datetimes):
    datetime = Datetime()
    for series in datetimes:
        assert str(series.dtype) == "object"
        transform = datetime.transform(series)
        assert str(transform.dtype) == "datetime64[ns]"
        assert datetime.datetime_format is not None


def test_datetime_inference_ambiguous_format():
    datetime = Datetime()
    dates = pd.Series(["01/01/2017"] * 2 + ["13/12/2017"], name="dates")
    warning = (
        "Some rows in series 'dates' are incompatible with datetime format "
        "'%m/%d/%Y' and have been replaced with null values. You may be able "
        "to fix this by using an instantiated Datetime logical type with a different "
        "format string specified for this column during Woodwork initialization."
    )
    with pytest.warns(TypeConversionWarning, match=warning):
        transformed = datetime.transform(dates)
    assert str(transformed.dtype) == "datetime64[ns]"
    assert transformed[2] is pd.NaT
    assert datetime.datetime_format == "%m/%d/%Y"


def test_datetime_coerce_user_format():
    datetime = Datetime(datetime_format="%m/%d/%Y")
    dates = pd.Series(["01/01/2017"] * 2 + ["13/12/2017"], name="dates")
    warning = (
        "Some rows in series 'dates' are incompatible with datetime format "
        "'%m/%d/%Y' and have been replaced with null values. You may be able "
        "to fix this by using an instantiated Datetime logical type with a different "
        "format string specified for this column during Woodwork initialization."
    )
    with pytest.warns(TypeConversionWarning, match=warning):
        transformed = datetime.transform(dates)
    assert str(transformed.dtype) == "datetime64[ns]"
    assert transformed[2] is pd.NaT
    assert datetime.datetime_format == "%m/%d/%Y"


def test_ordinal_transform(sample_series):
    series_type = str(type(sample_series))
    dask = "dask" in series_type
    spark = "spark" in series_type

    if dask or spark:
        pytest.xfail(
            "Fails with Dask and Spark - ordinal data validation not supported"
        )

    ordinal_incomplete_order = Ordinal(order=["a", "b"])
    error_msg = re.escape(
        "Ordinal column sample_series contains values that are not "
        "present in the order values provided: ['c']"
    )

    with pytest.raises(ValueError, match=error_msg):
        ordinal_incomplete_order.transform(sample_series)


def test_ordinal_validate(sample_series):
    series_type = str(type(sample_series))
    dask = "dask" in series_type
    spark = "spark" in series_type

    if dask or spark:
        pytest.xfail(
            "Fails with Dask and Spark - ordinal data validation not supported"
        )

    ordinal_incomplete_order = Ordinal(order=["a", "b"])
    error_msg = re.escape(
        "Ordinal column sample_series contains values that are not "
        "present in the order values provided: ['c']"
    )

    with pytest.raises(ValueError, match=error_msg):
        ordinal_incomplete_order.validate(sample_series)

    new_type = "string"
    error_message = re.escape(
        f"Series dtype '{new_type}' is incompatible with ordinal dtype."
    )
    with pytest.raises(TypeValidationError, match=error_message):
        ordinal_incomplete_order.validate(sample_series.astype(new_type))


def test_email_address_validate(sample_df):
    email_address = EmailAddress()
    dtype = email_address.primary_dtype
    series = sample_df["email"].astype(dtype)
    invalid_row = pd.Series({4: "bad_email"}, name="email").astype(dtype)

    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)

    assert email_address.validate(series) is None

    series = series.append(invalid_row).astype(dtype)
    match = "Series email contains invalid email address values. "
    match += "The email_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        email_address.validate(series)

    actual = email_address.validate(series, return_invalid_values=True)
    expected = pd.Series({4: "bad_email"}, name="email").astype(dtype)
    assert to_pandas(actual).equals(expected)


def test_url_validate(sample_df):
    logical_type = URL()
    dtype = logical_type.primary_dtype
    series = sample_df["url"].astype(dtype)
    invalid_row = pd.Series({4: "bad_url"}, name="url").astype(dtype)
    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)

    assert logical_type.validate(series) is None

    series = series.append(invalid_row).astype(dtype)
    match = "Series url contains invalid url values. "
    match += "The url_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        logical_type.validate(series)

    actual = logical_type.validate(series, return_invalid_values=True)
    expected = pd.Series({4: "bad_url"}, name="url").astype(dtype)
    assert to_pandas(actual).equals(expected)


@pytest.mark.parametrize(
    argnames="logical_type",
    ids=["age", "age_fractional", "age_nullable"],
    argvalues=[Age(), AgeFractional(), AgeNullable()],
)
def test_age_validate(sample_df, logical_type):
    series = sample_df["age"]
    if isinstance(logical_type, Age):
        series = series.dropna()

    dtype = logical_type.primary_dtype
    series = series.astype(dtype)

    assert logical_type.validate(series, return_invalid_values=False) is None
    invalid_row = pd.Series({4: -3}, name="age", dtype=dtype)

    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)
    series = series.append(invalid_row).astype(dtype)

    match = "Series age contains negative values."
    with pytest.raises(TypeValidationError, match=match):
        logical_type.validate(series, return_invalid_values=False)

    actual = logical_type.validate(series, return_invalid_values=True)
    assert to_pandas(actual).equals(to_pandas(invalid_row))


def test_phone_number_validate(sample_df):
    phone_number = PhoneNumber()
    dtype = phone_number.primary_dtype
    series = sample_df["phone_number"].astype(dtype)
    invalid_row = pd.Series({4: "bad_phone"}, name="phone_number").astype(dtype)

    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)

    assert phone_number.validate(series) is None

    series = series.append(invalid_row).astype(dtype)
    match = "Series phone_number contains invalid phone number values. "
    match += "The phone_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        phone_number.validate(series)

    actual = phone_number.validate(series, return_invalid_values=True)
    expected = pd.Series({4: "bad_phone"}, name="phone_number").astype(dtype)
    assert to_pandas(actual).equals(expected)


def test_phone_number_validate_complex(sample_df_phone_numbers):
    phone_number = PhoneNumber()
    dtype = phone_number.primary_dtype
    series = sample_df_phone_numbers["phone_number"].astype(dtype)
    # Current inference function does not match lack of area code
    invalid_row = pd.Series(
        {17: "252 9384", 18: "+1 194 129 1991", 19: "+01 236 248 8482"},
        name="phone_number",
    ).astype(dtype)

    series = series.append(invalid_row).astype(dtype)
    actual = phone_number.validate(series, return_invalid_values=True)
    expected = pd.Series(
        {17: "252 9384", 18: "+1 194 129 1991", 19: "+01 236 248 8482"},
        name="phone_number",
    ).astype(dtype)
    assert to_pandas(actual).equals(expected)


def test_postal_code_validate(sample_df_postal_code):
    pc = PostalCode()
    series = sample_df_postal_code["postal_code"]
    invalid_types = pd.Series(
        [
            "hello",
            "HELLO",
            "51342-HEL0",
        ]
    )
    series = series.append(invalid_types)
    series.name = "postal_code"
    match = "Series postal_code contains invalid postal code values. "
    match += "The postal_code_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        pc.validate(series)


def test_postal_code_validate_complex(sample_df_postal_code):
    pc = PostalCode()
    series = sample_df_postal_code["postal_code"]
    invalid_types = pd.Series(
        [
            "1234",
            "123455",
            "123456789",
            "1234-65433",
            "K1A0B1",  # Canadian formatting
            "K1A 0B1",
            "K1A-0B1",
            "DT1 1AE",  # UK formatting
            "DT1-1AE",
            "DT11AE",
        ]
    )
    actual = pc.validate(series, return_invalid_values=True)
    assert not len(actual)
    series = series.append(invalid_types)
    actual = pc.validate(series, return_invalid_values=True)
    pd.testing.assert_series_equal(actual, invalid_types)



from woodwork.deserializers.deserializer_base import Deserializer


class OrcDeserializer(Deserializer):
    """Deserialize Woodwork table from serialized data in orc format."""

    format = "orc"

    def read_from_local_path(self):
        lib = self._get_library()
        return lib.read_orc(self.read_path)



import pandas as pd

from ._constants import ESTIMATED_COLUMN_NAME, OBSERVED_COLUMN_NAME
from ._get_ranges import _get_ranges
from ._types import RangeObject


def _determine_extra_values(estimated_ts, observed_ts):
    """Calculate Extra Values in time_series:

    Args:
        estimated_ts (Series): The estimated time_series.
        observed_ts (Series): The observed time_series.

    Returns:
        (list(RangeObject)): A list of RangeObject data objects. A RangeObject has the following properties:

        - dt: an ISO 8601 formatted string of the first extra timestamp
        - idx: first index of the extra timestamp. Index is relative to observed timeseries
        - range: the number of sequential elements that are extra
    """
    estimated_df = pd.DataFrame({ESTIMATED_COLUMN_NAME: estimated_ts})
    observed_df = pd.DataFrame({OBSERVED_COLUMN_NAME: observed_ts})

    merged_df = estimated_df.merge(
        observed_df,
        how="right",
        left_on=ESTIMATED_COLUMN_NAME,
        right_on=OBSERVED_COLUMN_NAME,
    )

    estimated_null = merged_df[merged_df[ESTIMATED_COLUMN_NAME].isnull()]

    if len(estimated_null) == 0:
        return []

    ranges = _get_ranges(estimated_null.index)

    return [
        RangeObject(
            dt=observed_ts.iloc[start_idx].isoformat(),
            idx=observed_ts.index[start_idx],
            range=end_idx - start_idx + 1,
        )
        for start_idx, end_idx in ranges
    ]



import numpy as np
import pandas as pd
import pytest

from woodwork.statistics_utils import infer_frequency
from woodwork.tests.fixtures.datetime_freq import (
    generate_infer_error_messages,
    generate_pandas_inferrable,
    inferrable_freq_fixtures,
)


@pytest.mark.parametrize("freq", ["H", "2D", "3W-FRI", "M", "B", "Q-NOV"])
@pytest.mark.parametrize("error_range", [1, 2, 10])
def test_inferable_temporal_frequencies_missing(freq, error_range):
    # strip off first element, since it probably doesn't agree with freq
    dates = (
        (pd.date_range("2005-01-01", periods=1001, freq=freq)[1:])
        .to_series()
        .reset_index(drop=True)
    )

    actual_range_start = dates.loc[0].isoformat()
    actual_range_end = dates.loc[len(dates) - 1].isoformat()

    idx = len(dates) // 2

    d = dates[idx].isoformat()

    dates_observed = dates.drop(dates.loc[idx : idx + error_range - 1].index)
    dates_observed = dates_observed.reset_index(drop=True)

    expected_debug_obj = {
        "actual_range_start": actual_range_start,
        "actual_range_end": actual_range_end,
        "message": None,
        "estimated_freq": freq,
        "estimated_range_start": dates_observed.loc[0].isoformat(),
        "estimated_range_end": dates_observed.loc[len(dates_observed) - 1].isoformat(),
        "duplicate_values": [],
        "missing_values": [
            {"dt": d, "idx": idx, "range": error_range},
        ],
        "extra_values": [],
        "nan_values": [],
    }

    inferred_freq, actual_debug_obj = infer_frequency(
        observed_ts=dates_observed, debug=True
    )

    assert inferred_freq is None
    assert actual_debug_obj == expected_debug_obj


@pytest.mark.parametrize("freq", ["H", "2D", "3W-FRI", "M", "B", "Q-NOV"])
@pytest.mark.parametrize("error_range", [1, 2, 10])
def test_inferable_temporal_frequencies_duplicates(freq, error_range):
    # strip off first element, since it probably doesn't agree with freq
    dates = (
        (pd.date_range("2005-01-01", periods=1001, freq=freq)[1:])
        .to_series()
        .reset_index(drop=True)
    )

    actual_range_start = dates.loc[0].isoformat()
    actual_range_end = dates.loc[len(dates) - 1].isoformat()

    idx = len(dates) // 2

    d = dates[idx].isoformat()

    dates_observed = pd.concat(
        [
            dates[: idx + 1],
            pd.Series(np.full((error_range,), dates[idx])),
            dates[idx + 1 :],
        ]
    )
    dates_observed = dates_observed.reset_index(drop=True)

    expected_debug_obj = {
        "actual_range_start": actual_range_start,
        "actual_range_end": actual_range_end,
        "message": None,
        "estimated_freq": freq,
        "estimated_range_start": dates_observed.loc[0].isoformat(),
        "estimated_range_end": dates_observed.loc[len(dates_observed) - 1].isoformat(),
        "duplicate_values": [
            {"dt": d, "idx": idx + 1, "range": error_range},
        ],
        "missing_values": [],
        "extra_values": [],
        "nan_values": [],
    }

    inferred_freq, actual_debug_obj = infer_frequency(
        observed_ts=dates_observed, debug=True
    )

    assert inferred_freq is None
    assert actual_debug_obj == expected_debug_obj


@pytest.mark.parametrize("freq", ["H", "2D", "3W-FRI", "M", "B", "Q-NOV"])
@pytest.mark.parametrize("error_range", [1, 2, 10])
def test_inferable_temporal_frequencies_extra(freq, error_range):
    # strip off first element, since it probably doesn't agree with freq
    dates = (
        (pd.date_range("2005-01-01", periods=1001, freq=freq)[1:])
        .to_series()
        .reset_index(drop=True)
    )

    actual_range_start = dates.loc[0].isoformat()
    actual_range_end = dates.loc[len(dates) - 1].isoformat()

    idx = len(dates) // 2

    extra = (
        pd.date_range(dates[idx], periods=error_range + 1, freq="N")[1:]
    ).to_series()
    d = extra[0].isoformat()
    dates_observed = pd.concat([dates[: idx + 1], extra, dates[idx + 1 :]])
    dates_observed = dates_observed.reset_index(drop=True)

    expected_debug_obj = {
        "actual_range_start": actual_range_start,
        "actual_range_end": actual_range_end,
        "message": None,
        "estimated_freq": freq,
        "estimated_range_start": dates_observed.loc[0].isoformat(),
        "estimated_range_end": dates_observed.loc[len(dates_observed) - 1].isoformat(),
        "duplicate_values": [],
        "missing_values": [],
        "extra_values": [
            {"dt": d, "idx": idx + 1, "range": error_range},
        ],
        "nan_values": [],
    }

    inferred_freq, actual_debug_obj = infer_frequency(
        observed_ts=dates_observed, debug=True
    )

    assert inferred_freq is None
    assert actual_debug_obj == expected_debug_obj


@pytest.mark.parametrize("case", inferrable_freq_fixtures)
def test_inferable_temporal_frequencies_cases(case):
    input_series = case["dates"]

    expected_debug_obj = case["expected_debug_obj"]

    inferred_freq, actual_debug_obj = infer_frequency(
        observed_ts=input_series, debug=True
    )

    assert inferred_freq is None
    assert actual_debug_obj == expected_debug_obj


@pytest.mark.parametrize("case", generate_pandas_inferrable())
def test_pandas_inferable_temporal_frequencies(case):
    input_series = case["dates"]

    expected_infer_freq = case["expected_infer_freq"]

    inferred_freq = infer_frequency(observed_ts=input_series, debug=False)

    assert inferred_freq == expected_infer_freq


@pytest.mark.parametrize("case", generate_infer_error_messages())
def test_error_messages(case):
    input_series = case["dates"]

    expected_debug_obj = case["expected_debug_obj"]

    inferred_freq, actual_debug_obj = infer_frequency(
        observed_ts=input_series, debug=True
    )

    assert inferred_freq is None
    assert actual_debug_obj == expected_debug_obj



from woodwork.config import config
from woodwork.logical_types import Datetime, Timedelta
from woodwork.statistics_utils.frequency_inference._infer_frequency import (
    infer_frequency,
)


def _infer_temporal_frequencies(dataframe, temporal_columns=None, debug=False):
    """Infers the observation frequency (daily, biweekly, yearly, etc) of each temporal column
            in the DataFrame. Temporal columns are ones with the logical type Datetime or Timedelta.

    Args:
        dataframe (DataFrame): The DataFrame for which frequncies should be inferred.
        temporal_columns (list[str], optional): Columns for which frequencies should be inferred. Must be columns
            that are present in the DataFrame and are temporal in nature. Defaults to None. If not
            specified, all temporal columns will have their frequencies inferred.
        debug (boolean): A flag used to determine if more information should be returned for each temporal column if
                no uniform frequency was found. If True, a tuple is returned for each temporal column.
    Returns:
        (dict): A dictionary where each key is a temporal column from the DataFrame, and the
            value is its observation frequency represented as a pandas offset alias string (D, M, Y, etc.)
            or None if no uniform frequency was present in the data. If Debug=True, the response is a tuple.
    """
    logical_types = dataframe.ww.logical_types

    if temporal_columns is None:
        temporal_columns = [
            col
            for col, ltype in logical_types.items()
            if isinstance(ltype, (Datetime, Timedelta))
        ]
    else:
        for col in temporal_columns:
            if col not in dataframe:
                raise ValueError(f"Column {col} not found in dataframe.")
            ltype = logical_types[col]
            if not isinstance(ltype, (Datetime, Timedelta)):
                raise TypeError(
                    f"Cannot determine frequency for column {col} with logical type {ltype}"
                )

    window_length = config.get_option("frequence_inference_window_length")
    threshold = config.get_option("frequence_inference_threshold")

    return {
        col: infer_frequency(
            observed_ts=dataframe[col],
            debug=debug,
            window_length=window_length,
            threshold=threshold,
        )
        for col in temporal_columns
    }



import os
from unittest.mock import patch

import pandas as pd
import pytest

import woodwork as ww
from woodwork.serializers.orc_serializer import save_orc_file
from woodwork.utils import _replace_nan_strings


def test_read_file_errors_no_content_type(sample_df_pandas, tmpdir):
    filepath = os.path.join(tmpdir, "sample")
    sample_df_pandas.to_csv(filepath, index=False)

    no_type_error = (
        "Content type could not be inferred. Please specify content_type and try again."
    )
    with pytest.raises(RuntimeError, match=no_type_error):
        ww.read_file(filepath=filepath)


def test_read_file_errors_unsupported(sample_df_pandas, tmpdir):
    filepath = os.path.join(tmpdir, "sample")
    sample_df_pandas.to_feather(filepath)

    content_type = "notacontenttype"
    not_supported_error = (
        "Reading from content type {} is not currently supported".format(content_type)
    )
    with pytest.raises(RuntimeError, match=not_supported_error):
        ww.read_file(filepath=filepath, content_type=content_type)


@patch("woodwork.table_accessor._validate_accessor_params")
def test_read_file_validation_control(
    mock_validate_accessor_params, sample_df_pandas, tmpdir
):
    filepath = os.path.join(tmpdir, "sample.csv")
    sample_df_pandas.to_csv(filepath, index=False)

    assert not mock_validate_accessor_params.called
    ww.read_file(filepath=filepath, validate=False)
    assert not mock_validate_accessor_params.called
    ww.read_file(filepath=filepath)
    assert mock_validate_accessor_params.called


@pytest.mark.parametrize(
    "filepath, exportfn, kwargs, pandas_nullable_fix",
    [
        ("sample.csv", ("to_csv", {"index": False}), {}, True),
        ("sample.csv", ("to_csv", {"index": False}), {"content_type": "csv"}, True),
        (
            "sample.csv",
            ("to_csv", {"index": False}),
            {
                "index": "id",
                "time_index": "signup_date",
                "logical_types": {
                    "full_name": "NaturalLanguage",
                    "phone_number": "PhoneNumber",
                    "is_registered": "BooleanNullable",
                    "age": "IntegerNullable",
                    "nullable_integer": "IntegerNullable",
                },
                "semantic_tags": {
                    "age": ["tag1", "tag2"],
                    "is_registered": ["tag3", "tag4"],
                },
                "column_origins": {
                    "full_name": "csv",
                    "phone_number": "base",
                    "is_registered": "engineered",
                },
            },
            False,
        ),
        (
            "sample.csv",
            ("to_csv", {"index": False}),
            {
                "nrows": 2,
                "dtype": {
                    "age": "Int64",
                    "is_registered": "boolean",
                    "nullable_integer": "Int64",
                },
            },
            False,
        ),
        ("sample.feather", ("to_feather", {}), {}, False),
        (
            "sample.feather",
            ("to_feather", {}),
            {"content_type": "feather", "index": "id"},
            False,
        ),
        (
            "sample.feather",
            ("to_feather", {}),
            {"content_type": "application/feather", "index": "id"},
            False,
        ),
        ("sample.arrow", ("to_feather", {}), {}, False),
        (
            "sample.arrow",
            ("to_feather", {}),
            {"content_type": "arrow", "index": "id"},
            False,
        ),
        (
            "sample.arrow",
            ("to_feather", {}),
            {"content_type": "application/arrow", "index": "id"},
            False,
        ),
        ("sample.parquet", ("to_parquet", {}), {}, False),
        (
            "sample.parquet",
            ("to_parquet", {}),
            {"content_type": "parquet", "index": "id", "use_nullable_dtypes": True},
            False,
        ),
        (
            "sample.parquet",
            ("to_parquet", {}),
            {
                "content_type": "application/parquet",
                "index": "id",
                "use_nullable_dtypes": True,
            },
            False,
        ),
        ("sample.orc", (save_orc_file, {}), {}, True),
        (
            "sample.orc",
            (save_orc_file, {}),
            {"content_type": "orc", "index": "id"},
            True,
        ),
        (
            "sample.orc",
            (save_orc_file, {}),
            {"content_type": "application/orc", "index": "id"},
            True,
        ),
    ],
)
def test_read_file(
    sample_df_pandas, tmpdir, filepath, exportfn, kwargs, pandas_nullable_fix
):
    filepath = os.path.join(tmpdir, filepath)
    func, func_kwargs = exportfn
    if isinstance(func, str):
        getattr(sample_df_pandas, func)(filepath, **func_kwargs)
    else:
        # Call save_orc_file to save orc data since pandas does not have a to_orc method
        func(sample_df_pandas, filepath, **func_kwargs)
    df = ww.read_file(filepath=filepath, **kwargs)
    assert isinstance(df.ww.schema, ww.table_schema.TableSchema)

    schema_df = sample_df_pandas.copy()
    if pandas_nullable_fix:
        # pandas does not read data into nullable types currently from csv or orc,
        # so the types in df will be different than the types inferred from sample_df_pandas
        # which uses the nullable types
        schema_df = schema_df.astype(
            {"age": "float64", "nullable_integer": "float64", "is_registered": "object"}
        )

    if func in [
        "to_csv",
        save_orc_file,
    ]:  # categorical column not inferred as categorical
        schema_df["categorical"] = schema_df["categorical"].astype("string")

    schema_df.ww.init(
        index=kwargs.get("index"),
        time_index=kwargs.get("time_index"),
        logical_types=kwargs.get("logical_types"),
        semantic_tags=kwargs.get("semantic_tags"),
        column_origins=kwargs.get("column_origins"),
    )

    if func == "to_csv":
        for c in ["signup_date", "datetime_with_NaT"]:
            df.ww.logical_types[
                c
            ].datetime_format = None  # read_csv reads datetimes as strings and infers datetime during transform

    assert df.ww.schema == schema_df.ww.schema

    if "nrows" in kwargs:
        assert len(df) == kwargs["nrows"]
        schema_df = schema_df.head(kwargs["nrows"])

    pd.testing.assert_frame_equal(df, schema_df)


def test_replace_nan_strings():
    data = {
        "double": ["<NA>", "6.2", "4.2", "3.11"],
        "integer": ["<NA>", "6", "4", "3"],
        "null": ["<NA>", "", "nan", None],
        "null_string": pd.Series(["<NA>", "", "nan", ""], dtype="string"),
        "Int64": pd.Series([1, 2, 3, 4], dtype="Int64"),
        "Float64": pd.Series([1.1, 2.2, 3.3, 4.4], dtype="Float64"),
        "boolean": pd.Series([True, True, False, False], dtype="boolean"),
        "int64": pd.Series([1, 2, 3, 4], dtype="int64"),
        "double2": pd.Series([1, 2, 3, 4.5], dtype="float64"),
        "bool": pd.Series([True, True, False, False], dtype="bool"),
        "category": pd.Series([1, 2, 2, 1], dtype="category"),
    }

    expected_null_count = {
        "double": 1,
        "integer": 1,
        "null": 4,
        "null_string": 4,
        "Int64": 0,
        "Float64": 0,
        "boolean": 0,
        "int64": 0,
        "double2": 0,
        "bool": 0,
        "category": 0,
    }

    df = pd.DataFrame(data=data)
    replaced_df = _replace_nan_strings(df)
    for col in replaced_df:
        assert replaced_df[col].isnull().sum() == expected_null_count[col]


def test_replace_nan_strings_with_read_file(tmpdir):
    filepath = os.path.join(tmpdir, "data.parquet")
    content_type = "application/parquet"

    data = {
        "double": ["<NA>", "6.2", "4.2", "3.11"],
        "integer": ["<NA>", "6", "4", "3"],
        "null": ["<NA>", "", None, "nan"],
    }

    df = pd.DataFrame(data=data)
    df.to_parquet(filepath)

    # Without replacement
    actual = ww.read_file(
        content_type=content_type,
        filepath=filepath,
        replace_nan=False,
    )
    assert actual.isnull().sum().sum() == 1

    # With replacement
    actual = ww.read_file(
        content_type=content_type,
        filepath=filepath,
        replace_nan=True,
    )
    assert actual.isnull().sum().sum() == 6



import datetime
import json
import os
import tarfile
import tempfile

from woodwork.accessor_utils import _is_dask_dataframe, _is_spark_dataframe
from woodwork.exceptions import WoodworkFileExistsError
from woodwork.logical_types import LatLong
from woodwork.s3_utils import get_transport_params, use_smartopen
from woodwork.type_sys.utils import _get_ltype_class, _get_specified_ltype_params
from woodwork.utils import _is_s3, _is_url

SCHEMA_VERSION = "12.0.0"

PYARROW_IMPORT_ERROR_MESSAGE = (
    f"The pyarrow library is required to serialize to {format}.\n"
    "Install via pip:\n"
    "    pip install pyarrow\n"
    "Install via conda:\n"
    "   conda install pyarrow -c conda-forge"
)


class Serializer:
    def __init__(self, path, filename, data_subdirectory, typing_info_filename):
        self.path = path
        self.write_path = None
        self.filename = filename
        self.data_subdirectory = data_subdirectory
        self.typing_info_filename = typing_info_filename
        self.dataframe = None
        self.typing_info = None
        self.location = None
        self.kwargs = {}

    def serialize(self, dataframe, profile_name, **kwargs):
        """Serialize data and typing information to disk."""
        self.dataframe = dataframe
        self.typing_info = typing_info_to_dict(self.dataframe)

        if _is_s3(self.path):
            self.save_to_s3(profile_name)
        elif _is_url(self.path):
            raise ValueError("Writing to URLs is not supported")
        else:
            self.write_path = os.path.abspath(self.path)
            self.save_to_local_path()

    def save_to_local_path(self):
        """Serialize data and typing information to a local directory."""
        if self.data_subdirectory:
            location = os.path.join(self.write_path, self.data_subdirectory)
            os.makedirs(location, exist_ok=True)
        else:
            os.makedirs(self.write_path, exist_ok=True)
        self.write_dataframe()
        self.write_typing_info()

    def save_to_s3(self, profile_name):
        """Serialize data and typing information to S3."""
        with tempfile.TemporaryDirectory() as tmpdir:
            self.write_path = tmpdir
            self.save_to_local_path()
            archive_file_path = self._create_archive()
            transport_params = get_transport_params(profile_name)
            use_smartopen(
                archive_file_path,
                self.path,
                read=False,
                transport_params=transport_params,
            )

    def write_dataframe(self):
        """Save dataframe to disk."""
        raise NotImplementedError(
            "Must define write_dataframe on Serializer subclass"
        )  # pragma: no cover

    def write_typing_info(self):
        """Save Woodwork typing information JSON file to disk."""
        loading_info = {
            "location": self.location,
            "type": self.format,
            "params": self.kwargs,
        }
        self.typing_info["loading_info"].update(loading_info)
        file = os.path.join(self.write_path, self.typing_info_filename)

        if os.path.exists(file):
            message = f"Typing info already exists at '{file}'. "
            message += "Please remove or use a different filename."
            raise WoodworkFileExistsError(message)

        try:
            with open(file, "w") as file:
                json.dump(self.typing_info, file)
        except TypeError:
            raise TypeError(
                "Woodwork table is not json serializable. Check table and column metadata for values that may not be serializable."
            )

    def _get_filename(self):
        """Get the full filepath that should be used to save the data."""
        if self.filename is None:
            ww_name = self.dataframe.ww.name or "data"
            basename = ".".join([ww_name, self.format])
        else:
            basename = self.filename
        self.location = basename
        if self.data_subdirectory:
            self.location = os.path.join(self.data_subdirectory, basename)
        location = os.path.join(self.write_path, self.location)
        if os.path.exists(location):
            message = f"Data file already exists at '{location}'. "
            message += "Please remove or use a different filename."
            raise WoodworkFileExistsError(message)
        return location

    def _create_archive(self):
        """Create a tar archive of data and typing information."""
        file_name = "ww-{date:%Y-%m-%d_%H%M%S}.tar".format(date=datetime.datetime.now())
        file_path = os.path.join(self.write_path, file_name)
        tar = tarfile.open(str(file_path), "w")
        if self.typing_info_filename:
            tar.add(
                str(self.write_path) + f"/{self.typing_info_filename}",
                arcname=f"/{self.typing_info_filename}",
            )
        tar.add(
            str(self.write_path) + f"/{self.data_subdirectory}",
            arcname=f"/{self.data_subdirectory}",
        )
        tar.close()
        return file_path


def typing_info_to_dict(dataframe):
    """Creates the description for a Woodwork table, including typing information for each column
    and loading information.

    Args:
        dataframe (pd.DataFrame, dd.Dataframe, ks.DataFrame): DataFrame with Woodwork typing
            information initialized.

    Returns:
        dict: Dictionary containing Woodwork typing information
    """
    if _is_dask_dataframe(dataframe):
        # Need to determine the category info for Dask it can be saved below
        category_cols = [
            colname
            for colname, col in dataframe.ww._schema.columns.items()
            if col.is_categorical
        ]
        dataframe = dataframe.ww.categorize(columns=category_cols)
    ordered_columns = dataframe.columns

    def _get_physical_type_dict(column):
        type_dict = {"type": str(column.dtype)}
        if str(column.dtype) == "category":
            type_dict["cat_values"] = column.dtype.categories.to_list()
            type_dict["cat_dtype"] = str(column.dtype.categories.dtype)
        return type_dict

    column_typing_info = [
        {
            "name": col_name,
            "ordinal": ordered_columns.get_loc(col_name),
            "use_standard_tags": col.use_standard_tags,
            "logical_type": {
                "parameters": _get_specified_ltype_params(col.logical_type),
                "type": str(_get_ltype_class(col.logical_type)),
            },
            "physical_type": _get_physical_type_dict(dataframe[col_name]),
            "semantic_tags": sorted(list(col.semantic_tags)),
            "description": col.description,
            "origin": col.origin,
            "metadata": col.metadata,
        }
        for col_name, col in dataframe.ww.columns.items()
    ]

    if _is_dask_dataframe(dataframe):
        table_type = "dask"
    elif _is_spark_dataframe(dataframe):
        table_type = "spark"
    else:
        table_type = "pandas"

    return {
        "schema_version": SCHEMA_VERSION,
        "name": dataframe.ww.name,
        "index": dataframe.ww.index,
        "time_index": dataframe.ww.time_index,
        "column_typing_info": column_typing_info,
        "loading_info": {"table_type": table_type},
        "table_metadata": dataframe.ww.metadata,
    }


def clean_latlong(dataframe):
    """Convert latlong tuples to strings for parquet, arrow and feather file format.
    Attempting to serialize with tuples present results in an error"""
    latlong_columns = [
        col_name
        for col_name, col in dataframe.ww.columns.items()
        if _get_ltype_class(col.logical_type) == LatLong
    ]
    if len(latlong_columns) > 0:
        dataframe = dataframe.ww.copy()
        dataframe[latlong_columns] = dataframe[latlong_columns].astype(str)

    return dataframe



import pandas as pd
from dateutil.parser import ParserError

import woodwork as ww
from woodwork.accessor_utils import _is_dask_series, _is_spark_series
from woodwork.utils import import_or_none

ps = import_or_none("pyspark.pandas")
dd = import_or_none("dask.dataframe")


def col_is_datetime(col, datetime_format=None):
    """Determine if a dataframe column contains datetime values or not. Returns True if column
    contains datetimes, False if not. Optionally specify the datetime format string for the column.
    Will not infer numeric data as datetime."""
    if _is_spark_series(col):
        col = col.to_pandas()

    if pd.api.types.is_datetime64_any_dtype(col):
        return True

    col = col.dropna()
    if len(col) == 0:
        return False

    try:
        if pd.api.types.is_numeric_dtype(pd.Series(col.values.tolist())):
            return False
    except AttributeError:
        # given our current minimum dependencies (pandas 1.3.0 and pyarrow 4.0.1)
        # if we have a string dtype series, calling `col.values.tolist()` throws an error
        # AttributeError: 'StringArray' object has no attribute 'tolist'
        # this try/except block handles that, among other potential issues for experimental dtypes
        # until we find a more appropriate solution
        pass

    col = col.astype(str)

    try:
        pd.to_datetime(
            col,
            errors="raise",
            format=datetime_format,
            infer_datetime_format=True,
        )
        return True

    except (ParserError, ValueError, OverflowError, TypeError):
        return False


def _is_numeric_series(series, logical_type):
    """Determines whether a series will be considered numeric
    for the purposes of determining if it can be a time_index."""
    if _is_spark_series(series):
        series = series.to_pandas()
    if _is_dask_series(series):
        series = series.get_partition(0).compute()

    # If column can't be made to be numeric, don't bother checking Logical Type
    try:
        pd.to_numeric(series, errors="raise")
    except (ValueError, TypeError):
        return False

    if logical_type is not None:
        if isinstance(logical_type, str):
            logical_type = ww.type_system.str_to_logical_type(logical_type)

        # Allow numeric columns to be interpreted as Datetimes - doesn't allow strings even if they could be numeric
        if _get_ltype_class(
            logical_type
        ) == ww.logical_types.Datetime and pd.api.types.is_numeric_dtype(series):
            return True
    else:
        logical_type = ww.type_system.infer_logical_type(series)

    return "numeric" in logical_type.standard_tags


def list_logical_types():
    """Returns a dataframe describing all of the available Logical Types.

    Args:
        None

    Returns:
        pd.DataFrame: A dataframe containing details on each LogicalType, including
        the corresponding physical type and any standard semantic tags.
    """
    ltypes_df = pd.DataFrame(
        [
            {
                "name": ltype.__name__,
                "type_string": ltype.type_string,
                "description": ltype.__doc__,
                "physical_type": ltype.primary_dtype,
                "standard_tags": ltype.standard_tags,
                "is_default_type": ltype in ww.type_system._default_inference_functions,
                "is_registered": ltype in ww.type_system.registered_types,
                "parent_type": ww.type_system._get_parent(ltype),
            }
            for ltype in ww.logical_types.LogicalType.__subclasses__()
        ]
    )
    return ltypes_df.sort_values("name").reset_index(drop=True)


def list_semantic_tags():
    """Returns a dataframe describing all of the common semantic tags.

    Args:
        None

    Returns:
        pd.DataFrame: A dataframe containing details on each Semantic Tag, including
        the corresponding logical type(s).
    """
    sem_tags = {}
    for ltype in ww.type_system.registered_types:
        for tag in ltype.standard_tags:
            if tag in sem_tags:
                sem_tags[tag].append(ltype)
            else:
                sem_tags[tag] = [ltype]
    tags_df = pd.DataFrame(
        [
            {"name": tag, "is_standard_tag": True, "valid_logical_types": sem_tags[tag]}
            for tag in sem_tags
        ]
    )
    tags_df = tags_df.append(
        pd.DataFrame(
            [
                ["index", False, "Any LogicalType"],
                [
                    "time_index",
                    False,
                    [ww.type_system.str_to_logical_type("datetime")]
                    + sem_tags["numeric"],
                ],
                [
                    "date_of_birth",
                    False,
                    [ww.type_system.str_to_logical_type("datetime")],
                ],
                ["ignore", False, "Any LogicalType"],
                ["passthrough", False, "Any LogicalType"],
            ],
            columns=tags_df.columns,
        ),
        ignore_index=True,
    )
    return tags_df


def _get_ltype_class(ltype):
    if ltype in ww.type_system.registered_types:
        return ltype
    return ltype.__class__


def _get_specified_ltype_params(ltype):
    """Gets a dictionary of a LogicalType's parameters.

    Note: If the logical type has not been instantiated, no parameters have
    been specified for the LogicalType, so no parameters will be returned
    even if that LogicalType can have parameters set.

    Args:
        ltype (LogicalType): An instantiated or uninstantiated LogicalType

    Returns:
        dict: The LogicalType's specified parameters.
    """
    if ltype in ww.type_system.registered_types:
        # Do not reveal parameters for an uninstantiated LogicalType
        return {}
    return ltype.__dict__


def _is_categorical_series(series: pd.Series, threshold: float) -> bool:
    """
    Return ``True`` if the given series is "likely" to be categorical.
    Otherwise, return ``False``.  We say that a series is "likely" to be
    categorical if the percentage of unique values relative to total non-NA
    values is below a certain threshold.  In other words, if all values in the
    series are accounted for by a sufficiently small collection of unique
    values, then the series is categorical.
    """
    try:
        nunique = series.nunique()
    except TypeError as e:
        # It doesn't seem like there's a more elegant way to do this.  Pandas
        # doesn't provide an API that would give you any indication ahead of
        # time if a series with object dtype has any unhashable elements.
        if "unhashable type" in e.args[0]:
            return False
        else:
            raise  # pragma: no cover
    if nunique == 0:
        return False

    pct_unique = nunique / series.count()
    return pct_unique <= threshold
